from http import HTTPStatus
from typing import TYPE_CHECKING

import attr
from typing_extensions import TypeAlias

from synapse.api.errors import Codes, NotFoundError, SynapseError
from synapse.http.server import HttpServer
from synapse.http.servlet import (
    RestServlet,
    parse_and_validate_json_object_from_request,
    parse_integer,
    parse_string,
)
from synapse.http.site import SynapseRequest
from synapse.rest.client._base import client_patterns
from synapse.types import (
    JsonDict,
    RoomID,
    SlidingSyncStreamToken,
    ThreadSubscriptionsToken,
)
from synapse.types.handlers.sliding_sync import SlidingSyncResult
from synapse.types.rest import RequestBodyModel
from synapse.util.pydantic_models import AnyEventId

if TYPE_CHECKING:
    from synapse.server import HomeServer

_ThreadSubscription: TypeAlias = (
    SlidingSyncResult.Extensions.ThreadSubscriptionsExtension.ThreadSubscription
)
_ThreadUnsubscription: TypeAlias = (
    SlidingSyncResult.Extensions.ThreadSubscriptionsExtension.ThreadUnsubscription
)


class ThreadSubscriptionsRestServlet(RestServlet):
    PATTERNS = client_patterns(
        "/io.element.msc4306/rooms/(?P<room_id>[^/]*)/thread/(?P<thread_root_id>[^/]*)/subscription$",
        unstable=True,
        releases=(),
    )
    CATEGORY = "Thread Subscriptions requests (unstable)"

    def __init__(self, hs: "HomeServer"):
        self.auth = hs.get_auth()
        self.is_mine = hs.is_mine
        self.store = hs.get_datastores().main
        self.handler = hs.get_thread_subscriptions_handler()

    class PutBody(RequestBodyModel):
        automatic: AnyEventId | None = None
        """
        If supplied, the event ID of an event giving rise to this automatic subscription.

        If omitted, this subscription is a manual subscription.
        """

    async def on_GET(
        self, request: SynapseRequest, room_id: str, thread_root_id: str
    ) -> tuple[int, JsonDict]:
        RoomID.from_string(room_id)
        if not thread_root_id.startswith("$"):
            raise SynapseError(
                HTTPStatus.BAD_REQUEST, "Invalid event ID", errcode=Codes.INVALID_PARAM
            )
        requester = await self.auth.get_user_by_req(request)

        subscription = await self.handler.get_thread_subscription_settings(
            requester.user,
            room_id,
            thread_root_id,
        )

        if subscription is None:
            raise NotFoundError("Not subscribed.")

        return HTTPStatus.OK, {"automatic": subscription.automatic}

    async def on_PUT(
        self, request: SynapseRequest, room_id: str, thread_root_id: str
    ) -> tuple[int, JsonDict]:
        RoomID.from_string(room_id)
        if not thread_root_id.startswith("$"):
            raise SynapseError(
                HTTPStatus.BAD_REQUEST, "Invalid event ID", errcode=Codes.INVALID_PARAM
            )
        body = parse_and_validate_json_object_from_request(request, self.PutBody)

        requester = await self.auth.get_user_by_req(request)

        await self.handler.subscribe_user_to_thread(
            requester.user,
            room_id,
            thread_root_id,
            automatic_event_id=body.automatic,
        )

        return HTTPStatus.OK, {}

    async def on_DELETE(
        self, request: SynapseRequest, room_id: str, thread_root_id: str
    ) -> tuple[int, JsonDict]:
        RoomID.from_string(room_id)
        if not thread_root_id.startswith("$"):
            raise SynapseError(
                HTTPStatus.BAD_REQUEST, "Invalid event ID", errcode=Codes.INVALID_PARAM
            )
        requester = await self.auth.get_user_by_req(request)

        await self.handler.unsubscribe_user_from_thread(
            requester.user,
            room_id,
            thread_root_id,
        )

        return HTTPStatus.OK, {}


class ThreadSubscriptionsPaginationRestServlet(RestServlet):
    PATTERNS = client_patterns(
        "/io.element.msc4308/thread_subscriptions$",
        unstable=True,
        releases=(),
    )
    CATEGORY = "Thread Subscriptions requests (unstable)"

    # Maximum number of thread subscriptions to return in one request.
    MAX_LIMIT = 512

    def __init__(self, hs: "HomeServer"):
        self.auth = hs.get_auth()
        self.is_mine = hs.is_mine
        self.store = hs.get_datastores().main

    async def on_GET(self, request: SynapseRequest) -> tuple[int, JsonDict]:
        requester = await self.auth.get_user_by_req(request)

        limit = min(
            parse_integer(request, "limit", default=100, negative=False),
            ThreadSubscriptionsPaginationRestServlet.MAX_LIMIT,
        )
        from_end_opt = parse_string(request, "from", required=False)
        to_start_opt = parse_string(request, "to", required=False)
        _direction = parse_string(request, "dir", required=True, allowed_values=("b",))

        if limit <= 0:
            # condition needed because `negative=False` still allows 0
            raise SynapseError(
                HTTPStatus.BAD_REQUEST,
                "limit must be greater than 0",
                errcode=Codes.INVALID_PARAM,
            )

        if from_end_opt is not None:
            try:
                # because of backwards pagination, the `from` token is actually the
                # bound closest to the end of the stream
                end_stream_id = ThreadSubscriptionsToken.from_string(
                    from_end_opt
                ).stream_id
            except ValueError:
                raise SynapseError(
                    HTTPStatus.BAD_REQUEST,
                    "`from` is not a valid token",
                    errcode=Codes.INVALID_PARAM,
                )
        else:
            end_stream_id = self.store.get_max_thread_subscriptions_stream_id()

        if to_start_opt is not None:
            # because of backwards pagination, the `to` token is actually the
            # bound closest to the start of the stream
            try:
                start_stream_id = ThreadSubscriptionsToken.from_string(
                    to_start_opt
                ).stream_id
            except ValueError:
                # we also accept sliding sync `pos` tokens on this parameter
                try:
                    sliding_sync_pos = await SlidingSyncStreamToken.from_string(
                        self.store, to_start_opt
                    )
                    start_stream_id = (
                        sliding_sync_pos.stream_token.thread_subscriptions_key
                    )
                except ValueError:
                    raise SynapseError(
                        HTTPStatus.BAD_REQUEST,
                        "`to` is not a valid token",
                        errcode=Codes.INVALID_PARAM,
                    )
        else:
            # the start of time is ID 1; the lower bound is exclusive though
            start_stream_id = 0

        subscriptions = (
            await self.store.get_latest_updated_thread_subscriptions_for_user(
                requester.user.to_string(),
                from_id=start_stream_id,
                to_id=end_stream_id,
                limit=limit,
            )
        )

        subscribed_threads: dict[str, dict[str, JsonDict]] = {}
        unsubscribed_threads: dict[str, dict[str, JsonDict]] = {}
        for stream_id, room_id, thread_root_id, subscribed, automatic in subscriptions:
            if subscribed:
                subscribed_threads.setdefault(room_id, {})[thread_root_id] = (
                    attr.asdict(
                        _ThreadSubscription(
                            automatic=automatic,
                            bump_stamp=stream_id,
                        )
                    )
                )
            else:
                unsubscribed_threads.setdefault(room_id, {})[thread_root_id] = (
                    attr.asdict(_ThreadUnsubscription(bump_stamp=stream_id))
                )

        result: JsonDict = {}
        if subscribed_threads:
            result["subscribed"] = subscribed_threads
        if unsubscribed_threads:
            result["unsubscribed"] = unsubscribed_threads

        if len(subscriptions) == limit:
            # We hit the limit, so there might be more entries to return.
            # Generate a new token that has moved backwards, ready for the next
            # request.
            min_returned_stream_id, _, _, _, _ = subscriptions[0]
            result["end"] = ThreadSubscriptionsToken(
                # We subtract one because the 'later in the stream' bound is inclusive,
                # and we already saw the element at index 0.
                stream_id=min_returned_stream_id - 1
            ).to_string()

        return HTTPStatus.OK, result


def register_servlets(hs: "HomeServer", http_server: HttpServer) -> None:
    if hs.config.experimental.msc4306_enabled:
        ThreadSubscriptionsRestServlet(hs).register(http_server)
        ThreadSubscriptionsPaginationRestServlet(hs).register(http_server)



#
# This file is licensed under the Affero General Public License (AGPL) version 3.
#
# Copyright 2022 The Matrix.org Foundation C.I.C.
# Copyright (C) 2023 New Vector, Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# See the GNU Affero General Public License for more details:
# <https://www.gnu.org/licenses/agpl-3.0.html>.
#
# Originally licensed under the Apache License, Version 2.0:
# <http://www.apache.org/licenses/LICENSE-2.0>.
#
# [This file includes modifications made by New Vector Limited]
#
#
from typing import TYPE_CHECKING, Any

import attr

from synapse.api.errors import SynapseError, UnrecognizedRequestError
from synapse.push.clientformat import format_push_rules_for_user
from synapse.storage.push_rule import RuleNotFoundException
from synapse.synapse_rust.push import get_base_rule_ids
from synapse.types import JsonDict, StreamKeyType, UserID

if TYPE_CHECKING:
    from synapse.server import HomeServer


BASE_RULE_IDS = get_base_rule_ids()


@attr.s(slots=True, frozen=True, auto_attribs=True)
class RuleSpec:
    scope: str
    template: str
    rule_id: str
    attr: str | None


class PushRulesHandler:
    """A class to handle changes in push rules for users."""

    def __init__(self, hs: "HomeServer"):
        self._notifier = hs.get_notifier()
        self._main_store = hs.get_datastores().main

    async def set_rule_attr(
        self, user_id: str, spec: RuleSpec, val: bool | JsonDict
    ) -> None:
        """Set an attribute (enabled or actions) on an existing push rule.

        Notifies listeners (e.g. sync handler) of the change.

        Args:
            user_id: the user for which to modify the push rule.
            spec: the spec of the push rule to modify.
            val: the value to change the attribute to.

        Raises:
            RuleNotFoundException if the rule being modified doesn't exist.
            SynapseError(400) if the value is malformed.
            UnrecognizedRequestError if the attribute to change is unknown.
            InvalidRuleException if we're trying to change the actions on a rule but
                the provided actions aren't compliant with the spec.
        """
        if spec.attr not in ("enabled", "actions"):
            # for the sake of potential future expansion, shouldn't report
            # 404 in the case of an unknown request so check it corresponds to
            # a known attribute first.
            raise UnrecognizedRequestError()

        namespaced_rule_id = f"global/{spec.template}/{spec.rule_id}"
        rule_id = spec.rule_id
        is_default_rule = rule_id.startswith(".")
        if is_default_rule:
            if namespaced_rule_id not in BASE_RULE_IDS:
                raise RuleNotFoundException("Unknown rule %r" % (namespaced_rule_id,))
        if spec.attr == "enabled":
            if isinstance(val, dict) and "enabled" in val:
                val = val["enabled"]
            if not isinstance(val, bool):
                # Legacy fallback
                # This should *actually* take a dict, but many clients pass
                # bools directly, so let's not break them.
                raise SynapseError(400, "Value for 'enabled' must be boolean")
            await self._main_store.set_push_rule_enabled(
                user_id, namespaced_rule_id, val, is_default_rule
            )
        elif spec.attr == "actions":
            if not isinstance(val, dict):
                raise SynapseError(400, "Value must be a dict")
            actions = val.get("actions")
            if not isinstance(actions, list):
                raise SynapseError(400, "Value for 'actions' must be dict")
            check_actions(actions)
            rule_id = spec.rule_id
            is_default_rule = rule_id.startswith(".")
            if is_default_rule:
                if namespaced_rule_id not in BASE_RULE_IDS:
                    raise RuleNotFoundException(
                        "Unknown rule %r" % (namespaced_rule_id,)
                    )
            await self._main_store.set_push_rule_actions(
                user_id, namespaced_rule_id, actions, is_default_rule
            )
        else:
            raise UnrecognizedRequestError()

        self.notify_user(user_id)

    def notify_user(self, user_id: str) -> None:
        """Notify listeners about a push rule change.

        Args:
            user_id: the user ID the change is for.
        """
        stream_id = self._main_store.get_max_push_rules_stream_id()
        self._notifier.on_new_event(
            StreamKeyType.PUSH_RULES, stream_id, users=[user_id]
        )

    async def push_rules_for_user(
        self, user: UserID
    ) -> dict[str, dict[str, list[dict[str, Any]]]]:
        """
        Push rules aren't really account data, but get formatted as such for /sync.
        """
        user_id = user.to_string()
        rules_raw = await self._main_store.get_push_rules_for_user(user_id)
        rules = format_push_rules_for_user(user, rules_raw)
        return rules


def check_actions(actions: list[str | JsonDict]) -> None:
    """Check if the given actions are spec compliant.

    Args:
        actions: the actions to check.

    Raises:
        InvalidRuleException if the rules aren't compliant with the spec.
    """
    if not isinstance(actions, list):
        raise InvalidRuleException("No actions found")

    for a in actions:
        # "dont_notify" and "coalesce" are legacy actions. They are allowed, but
        # ignored (resulting in no action from the pusher).
        if a in ["notify", "dont_notify", "coalesce"]:
            pass
        elif isinstance(a, dict) and "set_tweak" in a:
            pass
        else:
            raise InvalidRuleException("Unrecognised action %s" % a)


class InvalidRuleException(Exception):
    pass



#
# This file is licensed under the Affero General Public License (AGPL) version 3.
#
#  Copyright 2022 The Matrix.org Foundation C.I.C.
# Copyright (C) 2023 New Vector, Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# See the GNU Affero General Public License for more details:
# <https://www.gnu.org/licenses/agpl-3.0.html>.
#
# Originally licensed under the Apache License, Version 2.0:
# <http://www.apache.org/licenses/LICENSE-2.0>.
#
# [This file includes modifications made by New Vector Limited]
#
#

"""
This module exposes a single function which checks synapse's dependencies are present
and correctly versioned. It makes use of `importlib.metadata` to do so. The details
are a bit murky: there's no easy way to get a map from "extras" to the packages they
require. But this is probably just symptomatic of Python's package management.
"""

import logging
from importlib import metadata
from typing import Any, Iterable, NamedTuple, Sequence, cast

from packaging.markers import Marker, Value, Variable, default_environment
from packaging.requirements import Requirement
from packaging.utils import canonicalize_name

DISTRIBUTION_NAME = "matrix-synapse"


__all__ = ["check_requirements"]

logger = logging.getLogger(__name__)


class DependencyException(Exception):
    @property
    def message(self) -> str:
        return "\n".join(
            [
                "Missing Requirements: %s" % (", ".join(self.dependencies),),
                "To install run:",
                "    pip install --upgrade --force %s" % (" ".join(self.dependencies),),
                "",
            ]
        )

    @property
    def dependencies(self) -> Iterable[str]:
        for i in self.args[0]:
            yield '"' + i + '"'


DEV_EXTRAS = {"lint", "mypy", "test", "dev"}
ALL_EXTRAS = metadata.metadata(DISTRIBUTION_NAME).get_all("Provides-Extra")
assert ALL_EXTRAS is not None
RUNTIME_EXTRAS = set(ALL_EXTRAS) - DEV_EXTRAS
VERSION = metadata.version(DISTRIBUTION_NAME)


def _marker_environment(extra: str) -> dict[str, str]:
    """Return the marker environment for `extra`, seeded with the current interpreter."""

    env = cast(dict[str, str], dict(default_environment()))
    env["extra"] = extra
    return env


def _is_dev_dependency(req: Requirement) -> bool:
    """Return True if `req` is a development dependency."""
    if req.marker is None:
        return False

    marker_extras = _extras_from_marker(req.marker)
    return any(
        extra in DEV_EXTRAS and req.marker.evaluate(_marker_environment(extra))
        for extra in marker_extras
    )


def _should_ignore_runtime_requirement(req: Requirement) -> bool:
    # This is a build-time dependency. Irritatingly, `poetry build` ignores the
    # requirements listed in the [build-system] section of pyproject.toml, so in order
    # to support `poetry install --without dev` we have to mark it as a runtime dependency.
    # See discussion on https://github.com/python-poetry/poetry/issues/6154 (it sounds
    # like the poetry authors don't consider this a bug?)
    #
    # In any case, workaround this by ignoring setuptools_rust here. (It might be
    # slightly cleaner to put `setuptools_rust` in a `build` extra or similar, but for
    # now let's do something quick and dirty.
    if canonicalize_name(req.name) == "setuptools-rust":
        return True
    return False


class Dependency(NamedTuple):
    requirement: Requirement
    must_be_installed: bool


def _generic_dependencies() -> Iterable[Dependency]:
    """Yield pairs (requirement, must_be_installed)."""
    requirements = metadata.requires(DISTRIBUTION_NAME)
    assert requirements is not None
    env_no_extra = _marker_environment("")
    for raw_requirement in requirements:
        req = Requirement(raw_requirement)
        if _is_dev_dependency(req) or _should_ignore_runtime_requirement(req):
            continue

        # https://packaging.pypa.io/en/latest/markers.html#usage notes that
        #   > Evaluating an extra marker with no environment is an error
        # so we pass in a dummy empty extra value here.
        must_be_installed = req.marker is None or req.marker.evaluate(env_no_extra)
        yield Dependency(req, must_be_installed)


def _dependencies_for_extra(extra: str) -> Iterable[Dependency]:
    """Yield additional dependencies needed for a given `extra`."""
    requirements = metadata.requires(DISTRIBUTION_NAME)
    assert requirements is not None
    env_no_extra = _marker_environment("")
    env_for_extra = _marker_environment(extra)
    for raw_requirement in requirements:
        req = Requirement(raw_requirement)
        if _is_dev_dependency(req):
            continue
        # Exclude mandatory deps by only selecting deps needed with this extra.
        if (
            req.marker is not None
            and req.marker.evaluate(env_for_extra)
            and not req.marker.evaluate(env_no_extra)
        ):
            yield Dependency(req, True)


def _values_from_marker_value(value: Value) -> set[str]:
    """Extract text values contained in a marker `Value`."""

    raw: Any = value.value
    if isinstance(raw, str):
        return {raw}
    if isinstance(raw, (tuple, list)):
        return {str(item) for item in raw}
    return {str(raw)}


def _extras_from_marker(marker: Marker | None) -> set[str]:
    """Return every `extra` referenced in the supplied marker tree."""

    extras: set[str] = set()

    if marker is None:
        return extras

    def collect(tree: object) -> None:
        if isinstance(tree, list):
            for item in tree:
                collect(item)
        elif isinstance(tree, tuple) and len(tree) == 3:
            lhs, _op, rhs = tree
            if (
                isinstance(lhs, Variable)
                and lhs.value == "extra"
                and isinstance(rhs, Value)
            ):
                extras.update(_values_from_marker_value(rhs))
            elif (
                isinstance(rhs, Variable)
                and rhs.value == "extra"
                and isinstance(lhs, Value)
            ):
                extras.update(_values_from_marker_value(lhs))

    collect(marker._markers)
    return extras


def _extras_to_consider_for_requirement(
    marker: Marker, base_candidates: Sequence[str]
) -> set[str]:
    """
    Augment `base_candidates` with extras explicitly mentioned in `marker`.

    Markers can mention extras (e.g. `extra == "saml2"`).
    """

    # Avoid modifying the input sequence.
    # Use a set to efficiently avoid duplicate extras.
    extras = set(base_candidates)

    for candidate in _extras_from_marker(marker):
        extras.add(candidate)

    return extras


def _marker_applies_for_any_extra(requirement: Requirement, extras: set[str]) -> bool:
    """Check whether a requirement's marker matches any evaluated `extra`."""

    if requirement.marker is None:
        return True

    return any(
        requirement.marker.evaluate(_marker_environment(extra)) for extra in extras
    )


def _not_installed(requirement: Requirement, extra: str | None = None) -> str:
    if extra:
        return (
            f"Synapse {VERSION} needs {requirement.name} for {extra}, "
            f"but it is not installed"
        )
    else:
        return f"Synapse {VERSION} needs {requirement.name}, but it is not installed"


def _incorrect_version(
    requirement: Requirement, got: str, extra: str | None = None
) -> str:
    if extra:
        return (
            f"Synapse {VERSION} needs {requirement} for {extra}, "
            f"but got {requirement.name}=={got}"
        )
    else:
        return (
            f"Synapse {VERSION} needs {requirement}, but got {requirement.name}=={got}"
        )


def _no_reported_version(requirement: Requirement, extra: str | None = None) -> str:
    if extra:
        return (
            f"Synapse {VERSION} needs {requirement} for {extra}, "
            f"but can't determine {requirement.name}'s version"
        )
    else:
        return (
            f"Synapse {VERSION} needs {requirement}, "
            f"but can't determine {requirement.name}'s version"
        )


def check_requirements(extra: str | None = None) -> None:
    """Check Synapse's dependencies are present and correctly versioned.

    If provided, `extra` must be the name of an packaging extra (e.g. "saml2" in
    `pip install matrix-synapse[saml2]`).

    If `extra` is None, this function checks that
    - all mandatory dependencies are installed and correctly versioned, and
    - each optional dependency that's installed is correctly versioned.

    If `extra` is not None, this function checks that
    - the dependencies needed for that extra are installed and correctly versioned.

    `marker`s are optional attributes on each requirement which specify
    conditions under which the requirement applies. For example, a requirement
    might only be needed on Windows, or with Python < 3.14. Markers can
    additionally mention `extras` themselves, meaning a requirement may not
    apply if the marker mentions an extra that the user has not asked for.

    This function skips a requirement when its markers do not apply in the
    current environment.

    :raises DependencyException: if a dependency is missing or incorrectly versioned.
    :raises ValueError: if this extra does not exist.
    """
    # First work out which dependencies are required, and which are optional.
    if extra is None:
        dependencies = _generic_dependencies()
    elif extra in RUNTIME_EXTRAS:
        dependencies = _dependencies_for_extra(extra)
    else:
        raise ValueError(f"Synapse {VERSION} does not provide the feature '{extra}'")

    deps_unfulfilled = []
    errors = []

    if extra is None:
        # Default to all mandatory dependencies (non-dev extras).
        # "" means all dependencies that aren't conditional on an extra.
        base_extra_candidates: Sequence[str] = ("", *RUNTIME_EXTRAS)
    else:
        base_extra_candidates = (extra,)

    for requirement, must_be_installed in dependencies:
        if requirement.marker is not None:
            candidate_extras = _extras_to_consider_for_requirement(
                requirement.marker, base_extra_candidates
            )
            # Skip checking this dependency if the requirement's marker object
            # (i.e. `python_version < "3.14" and os_name == "win32"`) does not
            # apply for any of the extras we're considering.
            if not _marker_applies_for_any_extra(requirement, candidate_extras):
                continue

        # Check if the requirement is installed and correctly versioned.
        try:
            dist: metadata.Distribution = metadata.distribution(requirement.name)
        except metadata.PackageNotFoundError:
            if must_be_installed:
                deps_unfulfilled.append(requirement.name)
                errors.append(_not_installed(requirement, extra))
        else:
            if dist.version is None:
                # This shouldn't happen---it suggests a borked virtualenv. (See
                # https://github.com/matrix-org/synapse/issues/12223)
                # Try to give a vaguely helpful error message anyway.
                # Type-ignore: the annotations don't reflect reality: see
                #     https://github.com/python/typeshed/issues/7513
                #     https://bugs.python.org/issue47060
                deps_unfulfilled.append(requirement.name)  # type: ignore[unreachable]
                errors.append(_no_reported_version(requirement, extra))

            # We specify prereleases=True to allow prereleases such as RCs.
            elif not requirement.specifier.contains(dist.version, prereleases=True):
                deps_unfulfilled.append(requirement.name)
                errors.append(_incorrect_version(requirement, dist.version, extra))

    if deps_unfulfilled:
        for err in errors:
            logger.error(err)

        raise DependencyException(deps_unfulfilled)



#
# This file is licensed under the Affero General Public License (AGPL) version 3.
#
# Copyright 2020 Sorunome
# Copyright 2018-2021 The Matrix.org Foundation C.I.C.
# Copyright (C) 2023 New Vector, Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# See the GNU Affero General Public License for more details:
# <https://www.gnu.org/licenses/agpl-3.0.html>.
#
# Originally licensed under the Apache License, Version 2.0:
# <http://www.apache.org/licenses/LICENSE-2.0>.
#
# [This file includes modifications made by New Vector Limited]
#
#
import logging
from collections import Counter
from typing import (
    TYPE_CHECKING,
    Any,
    Counter as CounterType,
    Iterable,
)

from synapse.api.constants import EventContentFields, EventTypes, Membership
from synapse.metrics import SERVER_NAME_LABEL, event_processing_positions
from synapse.storage.databases.main.state_deltas import StateDelta
from synapse.types import JsonDict
from synapse.util.duration import Duration
from synapse.util.events import get_plain_text_topic_from_event_content

if TYPE_CHECKING:
    from synapse.server import HomeServer

logger = logging.getLogger(__name__)


class StatsHandler:
    """Handles keeping the *_stats tables updated with a simple time-series of
    information about the users, rooms and media on the server, such that admins
    have some idea of who is consuming their resources.

    Heavily derived from UserDirectoryHandler
    """

    def __init__(self, hs: "HomeServer"):
        self.hs = hs
        self.server_name = hs.hostname
        self.store = hs.get_datastores().main
        self._storage_controllers = hs.get_storage_controllers()
        self.state = hs.get_state_handler()
        self.clock = hs.get_clock()
        self.notifier = hs.get_notifier()
        self.is_mine_id = hs.is_mine_id

        self.stats_enabled = hs.config.stats.stats_enabled

        # The current position in the current_state_delta stream
        self.pos: int | None = None

        # Guard to ensure we only process deltas one at a time
        self._is_processing = False

        if self.stats_enabled and hs.config.worker.run_background_tasks:
            self.notifier.add_replication_callback(self.notify_new_event)

            # We kick this off so that we don't have to wait for a change before
            # we start populating stats
            self.clock.call_later(
                Duration(seconds=0),
                self.notify_new_event,
            )

    def notify_new_event(self) -> None:
        """Called when there may be more deltas to process"""
        if not self.stats_enabled or self._is_processing:
            return

        self._is_processing = True

        async def process() -> None:
            try:
                await self._unsafe_process()
            finally:
                self._is_processing = False

        self.hs.run_as_background_process("stats.notify_new_event", process)

    async def _unsafe_process(self) -> None:
        # If self.pos is None then means we haven't fetched it from DB
        if self.pos is None:
            self.pos = await self.store.get_stats_positions()
            room_max_stream_ordering = self.store.get_room_max_stream_ordering()
            if self.pos > room_max_stream_ordering:
                # apparently, we've processed more events than exist in the database!
                # this can happen if events are removed with history purge or similar.
                logger.warning(
                    "Event stream ordering appears to have gone backwards (%i -> %i): "
                    "rewinding stats processor",
                    self.pos,
                    room_max_stream_ordering,
                )
                self.pos = room_max_stream_ordering

        # Loop round handling deltas until we're up to date

        while True:
            # Be sure to read the max stream_ordering *before* checking if there are any outstanding
            # deltas, since there is otherwise a chance that we could miss updates which arrive
            # after we check the deltas.
            room_max_stream_ordering = self.store.get_room_max_stream_ordering()
            if self.pos == room_max_stream_ordering:
                break

            logger.debug(
                "Processing room stats %s->%s", self.pos, room_max_stream_ordering
            )
            (
                max_pos,
                deltas,
            ) = await self._storage_controllers.state.get_current_state_deltas(
                self.pos, room_max_stream_ordering
            )

            if deltas:
                logger.debug("Handling %d state deltas", len(deltas))
                room_deltas, user_deltas = await self._handle_deltas(deltas)
            else:
                room_deltas = {}
                user_deltas = {}

            logger.debug("room_deltas: %s", room_deltas)
            logger.debug("user_deltas: %s", user_deltas)

            # Always call this so that we update the stats position.
            await self.store.bulk_update_stats_delta(
                self.clock.time_msec(),
                updates={"room": room_deltas, "user": user_deltas},
                stream_id=max_pos,
            )

            logger.debug("Handled room stats to %s -> %s", self.pos, max_pos)

            event_processing_positions.labels(
                name="stats", **{SERVER_NAME_LABEL: self.server_name}
            ).set(max_pos)

            self.pos = max_pos

    async def _handle_deltas(
        self, deltas: Iterable[StateDelta]
    ) -> tuple[dict[str, CounterType[str]], dict[str, CounterType[str]]]:
        """Called with the state deltas to process

        Returns:
            Two dicts: the room deltas and the user deltas,
            mapping from room/user ID to changes in the various fields.
        """

        room_to_stats_deltas: dict[str, CounterType[str]] = {}
        user_to_stats_deltas: dict[str, CounterType[str]] = {}

        room_to_state_updates: dict[str, dict[str, Any]] = {}

        for delta in deltas:
            logger.debug(
                "Handling: %r, %r %r, %s",
                delta.room_id,
                delta.event_type,
                delta.state_key,
                delta.event_id,
            )

            token = await self.store.get_earliest_token_for_stats("room", delta.room_id)

            # If the earliest token to begin from is larger than our current
            # stream ID, skip processing this delta.
            if token is not None and token >= delta.stream_id:
                logger.debug(
                    "Ignoring: %s as earlier than this room's initial ingestion event",
                    delta.event_id,
                )
                continue

            if delta.event_id is None and delta.prev_event_id is None:
                logger.error(
                    "event ID is None and so is the previous event ID. stream_id: %s",
                    delta.stream_id,
                )
                continue

            event_content: JsonDict = {}

            if delta.event_id is not None:
                event = await self.store.get_event(delta.event_id, allow_none=True)
                if event:
                    event_content = event.content or {}

            # All the values in this dict are deltas (RELATIVE changes)
            room_stats_delta = room_to_stats_deltas.setdefault(delta.room_id, Counter())

            room_state = room_to_state_updates.setdefault(delta.room_id, {})

            if delta.prev_event_id is None:
                # this state event doesn't overwrite another,
                # so it is a new effective/current state event
                room_stats_delta["current_state_events"] += 1

            if delta.event_type == EventTypes.Member:
                # we could use StateDeltasHandler._get_key_change here but it's
                # a bit inefficient given we're not testing for a specific
                # result; might as well just grab the prev_membership and
                # membership strings and compare them.
                # We take None rather than leave as a previous membership
                # in the absence of a previous event because we do not want to
                # reduce the leave count when a new-to-the-room user joins.
                prev_membership = None
                if delta.prev_event_id is not None:
                    prev_event = await self.store.get_event(
                        delta.prev_event_id, allow_none=True
                    )
                    if prev_event:
                        prev_event_content = prev_event.content
                        prev_membership = prev_event_content.get(
                            "membership", Membership.LEAVE
                        )

                membership = event_content.get("membership", Membership.LEAVE)

                if prev_membership is None:
                    logger.debug("No previous membership for this user.")
                elif membership == prev_membership:
                    pass  # noop
                elif prev_membership == Membership.JOIN:
                    room_stats_delta["joined_members"] -= 1
                elif prev_membership == Membership.INVITE:
                    room_stats_delta["invited_members"] -= 1
                elif prev_membership == Membership.LEAVE:
                    room_stats_delta["left_members"] -= 1
                elif prev_membership == Membership.BAN:
                    room_stats_delta["banned_members"] -= 1
                elif prev_membership == Membership.KNOCK:
                    room_stats_delta["knocked_members"] -= 1
                else:
                    raise ValueError(
                        "%r is not a valid prev_membership" % (prev_membership,)
                    )

                if membership == prev_membership:
                    pass  # noop
                elif membership == Membership.JOIN:
                    room_stats_delta["joined_members"] += 1
                elif membership == Membership.INVITE:
                    room_stats_delta["invited_members"] += 1
                elif membership == Membership.LEAVE:
                    room_stats_delta["left_members"] += 1
                elif membership == Membership.BAN:
                    room_stats_delta["banned_members"] += 1
                elif membership == Membership.KNOCK:
                    room_stats_delta["knocked_members"] += 1
                else:
                    raise ValueError("%r is not a valid membership" % (membership,))

                user_id = delta.state_key
                if self.is_mine_id(user_id):
                    # this accounts for transitions like leave → ban and so on.
                    has_changed_joinedness = (prev_membership == Membership.JOIN) != (
                        membership == Membership.JOIN
                    )

                    if has_changed_joinedness:
                        membership_delta = +1 if membership == Membership.JOIN else -1

                        user_to_stats_deltas.setdefault(user_id, Counter())[
                            "joined_rooms"
                        ] += membership_delta

                        room_stats_delta["local_users_in_room"] += membership_delta

            elif delta.event_type == EventTypes.Create:
                room_state["is_federatable"] = (
                    event_content.get(EventContentFields.FEDERATE, True) is True
                )
                room_type = event_content.get(EventContentFields.ROOM_TYPE)
                if isinstance(room_type, str):
                    room_state["room_type"] = room_type
            elif delta.event_type == EventTypes.JoinRules:
                room_state["join_rules"] = event_content.get("join_rule")
            elif delta.event_type == EventTypes.RoomHistoryVisibility:
                room_state["history_visibility"] = event_content.get(
                    "history_visibility"
                )
            elif delta.event_type == EventTypes.RoomEncryption:
                room_state["encryption"] = event_content.get(
                    EventContentFields.ENCRYPTION_ALGORITHM
                )
            elif delta.event_type == EventTypes.Name:
                room_state["name"] = event_content.get("name")
            elif delta.event_type == EventTypes.Topic:
                room_state["topic"] = get_plain_text_topic_from_event_content(
                    event_content
                )
            elif delta.event_type == EventTypes.RoomAvatar:
                room_state["avatar"] = event_content.get("url")
            elif delta.event_type == EventTypes.CanonicalAlias:
                room_state["canonical_alias"] = event_content.get("alias")
            elif delta.event_type == EventTypes.GuestAccess:
                room_state["guest_access"] = event_content.get(
                    EventContentFields.GUEST_ACCESS
                )

        for room_id, state in room_to_state_updates.items():
            logger.debug("Updating room_stats_state for %s: %s", room_id, state)
            await self.store.update_room_state(room_id, state)

        return room_to_stats_deltas, user_to_stats_deltas



#
# This file is licensed under the Affero General Public License (AGPL) version 3.
#
# Copyright 2020 The Matrix.org Foundation C.I.C.
# Copyright 2015, 2016 OpenMarket Ltd
# Copyright (C) 2023 New Vector, Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# See the GNU Affero General Public License for more details:
# <https://www.gnu.org/licenses/agpl-3.0.html>.
#
# Originally licensed under the Apache License, Version 2.0:
# <http://www.apache.org/licenses/LICENSE-2.0>.
#
# [This file includes modifications made by New Vector Limited]
#
#

import logging
import re
from collections import Counter
from http import HTTPStatus
from typing import TYPE_CHECKING, Any, Mapping

from pydantic import StrictBool, StrictStr, field_validator

from synapse.api.auth.mas import MasDelegatedAuth
from synapse.api.errors import (
    Codes,
    InteractiveAuthIncompleteError,
    InvalidAPICallError,
    SynapseError,
)
from synapse.http.server import HttpServer
from synapse.http.servlet import (
    RestServlet,
    parse_integer,
    parse_json_object_from_request,
    parse_string,
    validate_json_object,
)
from synapse.http.site import SynapseRequest
from synapse.logging.opentracing import set_tag
from synapse.rest.client._base import client_patterns, interactive_auth_handler
from synapse.types import JsonDict, StreamToken
from synapse.types.rest import RequestBodyModel
from synapse.util.cancellation import cancellable

if TYPE_CHECKING:
    from synapse.server import HomeServer

logger = logging.getLogger(__name__)


class KeyUploadServlet(RestServlet):
    """
    POST /keys/upload HTTP/1.1
    Content-Type: application/json

    {
        "device_keys": {
            "user_id": "<user_id>",
            "device_id": "<device_id>",
            "algorithms": [
                "m.olm.curve25519-aes-sha2",
            ]
            "keys": {
                "<algorithm>:<device_id>": "<key_base64>",
            },
            "signatures:" {
                "<user_id>" {
                    "<algorithm>:<device_id>": "<signature_base64>"
                }
            }
        },
        "fallback_keys": {
            "<algorithm>:<device_id>": "<key_base64>",
            "signed_<algorithm>:<device_id>": {
                "fallback": true,
                "key": "<key_base64>",
                "signatures": {
                    "<user_id>": {
                        "<algorithm>:<device_id>": "<key_base64>"
                    }
                }
            }
        }
        "one_time_keys": {
            "<algorithm>:<key_id>": "<key_base64>"
        },
    }

    response, e.g.:

    {
        "one_time_key_counts": {
            "curve25519": 10,
            "signed_curve25519": 20
        }
    }

    """

    PATTERNS = client_patterns("/keys/upload$")
    CATEGORY = "Encryption requests"

    def __init__(self, hs: "HomeServer"):
        super().__init__()
        self.auth = hs.get_auth()
        self.e2e_keys_handler = hs.get_e2e_keys_handler()
        self.device_handler = hs.get_device_handler()
        self._clock = hs.get_clock()
        self._store = hs.get_datastores().main

    class KeyUploadRequestBody(RequestBodyModel):
        """
        The body of a `POST /_matrix/client/v3/keys/upload` request.

        Based on https://spec.matrix.org/v1.16/client-server-api/#post_matrixclientv3keysupload.
        """

        class DeviceKeys(RequestBodyModel):
            algorithms: list[StrictStr]
            """The encryption algorithms supported by this device."""

            device_id: StrictStr
            """The ID of the device these keys belong to. Must match the device ID used when logging in."""

            keys: Mapping[StrictStr, StrictStr]
            """
            Public identity keys. The names of the properties should be in the
            format `<algorithm>:<device_id>`. The keys themselves should be encoded as
            specified by the key algorithm.
            """

            signatures: Mapping[StrictStr, Mapping[StrictStr, StrictStr]]
            """Signatures for the device key object. A map from user ID, to a map from "<algorithm>:<device_id>" to the signature."""

            user_id: StrictStr
            """The ID of the user the device belongs to. Must match the user ID used when logging in."""

        class KeyObject(RequestBodyModel):
            key: StrictStr
            """The key, encoded using unpadded base64."""

            fallback: StrictBool | None = False
            """Whether this is a fallback key. Only used when handling fallback keys."""

            signatures: Mapping[StrictStr, Mapping[StrictStr, StrictStr]]
            """Signature for the device. Mapped from user ID to another map of key signing identifier to the signature itself.

            See the following for more detail: https://spec.matrix.org/v1.16/appendices/#signing-details
            """

        device_keys: DeviceKeys | None = None
        """Identity keys for the device. May be absent if no new identity keys are required."""

        fallback_keys: Mapping[StrictStr, StrictStr | KeyObject] | None = None
        """
        The public key which should be used if the device's one-time keys are
        exhausted. The fallback key is not deleted once used, but should be
        replaced when additional one-time keys are being uploaded. The server
        will notify the client of the fallback key being used through `/sync`.

        There can only be at most one key per algorithm uploaded, and the server
        will only persist one key per algorithm.

        When uploading a signed key, an additional fallback: true key should be
        included to denote that the key is a fallback key.

        May be absent if a new fallback key is not required.
        """

        @field_validator("fallback_keys", mode="before")
        @classmethod
        def validate_fallback_keys(cls, v: Any) -> Any:
            if v is None:
                return v
            if not isinstance(v, dict):
                raise TypeError("fallback_keys must be a mapping")

            for k in v.keys():
                if not len(k.split(":")) == 2:
                    raise SynapseError(
                        code=HTTPStatus.BAD_REQUEST,
                        errcode=Codes.BAD_JSON,
                        msg=f"Invalid fallback_keys key {k!r}. "
                        'Expected "<algorithm>:<device_id>".',
                    )
            return v

        one_time_keys: Mapping[StrictStr, StrictStr | KeyObject] | None = None
        """
        One-time public keys for "pre-key" messages. The names of the properties
        should be in the format `<algorithm>:<key_id>`.

        The format of the key is determined by the key algorithm, see:
        https://spec.matrix.org/v1.16/client-server-api/#key-algorithms.
        """

        @field_validator("one_time_keys", mode="before")
        @classmethod
        def validate_one_time_keys(cls, v: Any) -> Any:
            if v is None:
                return v
            if not isinstance(v, dict):
                raise TypeError("one_time_keys must be a mapping")

            for k, _ in v.items():
                if not len(k.split(":")) == 2:
                    raise SynapseError(
                        code=HTTPStatus.BAD_REQUEST,
                        errcode=Codes.BAD_JSON,
                        msg=f"Invalid one_time_keys key {k!r}. "
                        'Expected "<algorithm>:<device_id>".',
                    )
            return v

    async def on_POST(self, request: SynapseRequest) -> tuple[int, JsonDict]:
        requester = await self.auth.get_user_by_req(request, allow_guest=True)
        user_id = requester.user.to_string()

        # Parse the request body. Validate separately, as the handler expects a
        # plain dict, rather than any parsed object.
        #
        # Note: It would be nice to work with a parsed object, but the handler
        # needs to encode portions of the request body as canonical JSON before
        # storing the result in the DB. There's little point in converted to a
        # parsed object and then back to a dict.
        body = parse_json_object_from_request(request)
        validate_json_object(body, self.KeyUploadRequestBody)

        device_id = requester.device_id

        if device_id is None:
            raise SynapseError(
                400, "To upload keys, you must pass device_id when authenticating"
            )

        if "device_keys" in body and isinstance(body["device_keys"], dict):
            # Validate the provided `user_id` and `device_id` fields in
            # `device_keys` match that of the requesting user. We can't do
            # this directly in the pydantic model as we don't have access
            # to the requester yet.
            #
            # TODO: We could use ValidationInfo when we switch to Pydantic v2.
            # https://docs.pydantic.dev/latest/concepts/validators/#validation-info
            if body["device_keys"].get("user_id") != user_id:
                raise SynapseError(
                    code=HTTPStatus.BAD_REQUEST,
                    errcode=Codes.BAD_JSON,
                    msg="Provided `user_id` in `device_keys` does not match that of the authenticated user",
                )
            if body["device_keys"].get("device_id") != device_id:
                raise SynapseError(
                    code=HTTPStatus.BAD_REQUEST,
                    errcode=Codes.BAD_JSON,
                    msg="Provided `device_id` in `device_keys` does not match that of the authenticated user device",
                )

        result = await self.e2e_keys_handler.upload_keys_for_user(
            user_id=user_id,
            device_id=device_id,
            keys=body,
        )

        return 200, result


class KeyQueryServlet(RestServlet):
    """
    POST /keys/query HTTP/1.1
    Content-Type: application/json
    {
      "device_keys": {
        "<user_id>": ["<device_id>"]
    } }

    HTTP/1.1 200 OK
    {
      "device_keys": {
        "<user_id>": {
          "<device_id>": {
            "user_id": "<user_id>", // Duplicated to be signed
            "device_id": "<device_id>", // Duplicated to be signed
            "valid_until_ts": <millisecond_timestamp>,
            "algorithms": [ // List of supported algorithms
              "m.olm.curve25519-aes-sha2",
            ],
            "keys": { // Must include a ed25519 signing key
              "<algorithm>:<key_id>": "<key_base64>",
            },
            "signatures:" {
              // Must be signed with device's ed25519 key
              "<user_id>/<device_id>": {
                "<algorithm>:<key_id>": "<signature_base64>"
              }
              // Must be signed by this server.
              "<server_name>": {
                "<algorithm>:<key_id>": "<signature_base64>"
    } } } } } }
    """

    PATTERNS = client_patterns("/keys/query$")
    CATEGORY = "Encryption requests"

    def __init__(self, hs: "HomeServer"):
        super().__init__()
        self.auth = hs.get_auth()
        self.e2e_keys_handler = hs.get_e2e_keys_handler()

    @cancellable
    async def on_POST(self, request: SynapseRequest) -> tuple[int, JsonDict]:
        requester = await self.auth.get_user_by_req(request, allow_guest=True)
        user_id = requester.user.to_string()
        device_id = requester.device_id
        timeout = parse_integer(request, "timeout", 10 * 1000)
        body = parse_json_object_from_request(request)

        device_keys = body.get("device_keys")
        if not isinstance(device_keys, dict):
            raise InvalidAPICallError("'device_keys' must be a JSON object")

        def is_list_of_strings(values: Any) -> bool:
            return isinstance(values, list) and all(isinstance(v, str) for v in values)

        if any(not is_list_of_strings(keys) for keys in device_keys.values()):
            raise InvalidAPICallError(
                "'device_keys' values must be a list of strings",
            )

        result = await self.e2e_keys_handler.query_devices(
            body, timeout, user_id, device_id
        )
        return 200, result


class KeyChangesServlet(RestServlet):
    """Returns the list of changes of keys between two stream tokens (may return
    spurious extra results, since we currently ignore the `to` param).

        GET /keys/changes?from=...&to=...

        200 OK
        { "changed": ["@foo:example.com"] }
    """

    PATTERNS = client_patterns("/keys/changes$")
    CATEGORY = "Encryption requests"

    def __init__(self, hs: "HomeServer"):
        super().__init__()
        self.auth = hs.get_auth()
        self.device_handler = hs.get_device_handler()
        self.store = hs.get_datastores().main

    @cancellable
    async def on_GET(self, request: SynapseRequest) -> tuple[int, JsonDict]:
        requester = await self.auth.get_user_by_req(request, allow_guest=True)

        from_token_string = parse_string(request, "from", required=True)
        set_tag("from", from_token_string)

        # We want to enforce they do pass us one, but we ignore it and return
        # changes after the "to" as well as before.
        #
        # XXX This does not enforce that "to" is passed.
        set_tag("to", str(parse_string(request, "to")))

        from_token = await StreamToken.from_string(self.store, from_token_string)

        user_id = requester.user.to_string()

        device_list_updates = await self.device_handler.get_user_ids_changed(
            user_id, from_token
        )

        response: JsonDict = {}
        response["changed"] = list(device_list_updates.changed)
        response["left"] = list(device_list_updates.left)

        return 200, response


class OneTimeKeyServlet(RestServlet):
    """
    POST /keys/claim HTTP/1.1
    {
      "one_time_keys": {
        "<user_id>": {
          "<device_id>": "<algorithm>"
    } } }

    HTTP/1.1 200 OK
    {
      "one_time_keys": {
        "<user_id>": {
          "<device_id>": {
            "<algorithm>:<key_id>": "<key_base64>"
    } } } }

    """

    PATTERNS = client_patterns("/keys/claim$")
    CATEGORY = "Encryption requests"

    def __init__(self, hs: "HomeServer"):
        super().__init__()
        self.auth = hs.get_auth()
        self.e2e_keys_handler = hs.get_e2e_keys_handler()

    async def on_POST(self, request: SynapseRequest) -> tuple[int, JsonDict]:
        requester = await self.auth.get_user_by_req(request, allow_guest=True)
        timeout = parse_integer(request, "timeout", 10 * 1000)
        body = parse_json_object_from_request(request)

        # Generate a count for each algorithm, which is hard-coded to 1.
        query: dict[str, dict[str, dict[str, int]]] = {}
        for user_id, one_time_keys in body.get("one_time_keys", {}).items():
            for device_id, algorithm in one_time_keys.items():
                query.setdefault(user_id, {})[device_id] = {algorithm: 1}

        result = await self.e2e_keys_handler.claim_one_time_keys(
            query, requester.user, timeout, always_include_fallback_keys=False
        )
        return 200, result


class UnstableOneTimeKeyServlet(RestServlet):
    """
    Identical to the stable endpoint (OneTimeKeyServlet) except it allows for
    querying for multiple OTKs at once and always includes fallback keys in the
    response.

    POST /keys/claim HTTP/1.1
    {
      "one_time_keys": {
        "<user_id>": {
          "<device_id>": ["<algorithm>", ...]
    } } }

    HTTP/1.1 200 OK
    {
      "one_time_keys": {
        "<user_id>": {
          "<device_id>": {
            "<algorithm>:<key_id>": "<key_base64>"
    } } } }

    """

    PATTERNS = [re.compile(r"^/_matrix/client/unstable/org.matrix.msc3983/keys/claim$")]
    CATEGORY = "Encryption requests"

    def __init__(self, hs: "HomeServer"):
        super().__init__()
        self.auth = hs.get_auth()
        self.e2e_keys_handler = hs.get_e2e_keys_handler()

    async def on_POST(self, request: SynapseRequest) -> tuple[int, JsonDict]:
        requester = await self.auth.get_user_by_req(request, allow_guest=True)
        timeout = parse_integer(request, "timeout", 10 * 1000)
        body = parse_json_object_from_request(request)

        # Generate a count for each algorithm.
        query: dict[str, dict[str, dict[str, int]]] = {}
        for user_id, one_time_keys in body.get("one_time_keys", {}).items():
            for device_id, algorithms in one_time_keys.items():
                query.setdefault(user_id, {})[device_id] = Counter(algorithms)

        result = await self.e2e_keys_handler.claim_one_time_keys(
            query, requester.user, timeout, always_include_fallback_keys=True
        )
        return 200, result


class SigningKeyUploadServlet(RestServlet):
    """
    POST /keys/device_signing/upload HTTP/1.1
    Content-Type: application/json

    {
    }
    """

    PATTERNS = client_patterns("/keys/device_signing/upload$", releases=("v3",))

    def __init__(self, hs: "HomeServer"):
        super().__init__()
        self.hs = hs
        self.auth = hs.get_auth()
        self.e2e_keys_handler = hs.get_e2e_keys_handler()
        self.auth_handler = hs.get_auth_handler()

    @interactive_auth_handler
    async def on_POST(self, request: SynapseRequest) -> tuple[int, JsonDict]:
        requester = await self.auth.get_user_by_req(request)
        user_id = requester.user.to_string()
        body = parse_json_object_from_request(request)

        (
            is_cross_signing_setup,
            master_key_updatable_without_uia,
        ) = await self.e2e_keys_handler.check_cross_signing_setup(user_id)

        # Resending exactly the same keys should just 200 OK without doing a UIA prompt.
        keys_are_different = await self.e2e_keys_handler.has_different_keys(
            user_id, body
        )
        if not keys_are_different:
            return 200, {}

        # The keys are different; is x-signing set up? If no, then this is first-time
        # setup, and that is allowed without UIA, per MSC3967.
        # If yes, then we need to authenticate the change.
        # MSC4190 can skip UIA for replacing cross-signing keys as well.
        if is_cross_signing_setup and not requester.app_service:
            # With MSC3861, UIA is not possible. Instead, the auth service has to
            # explicitly mark the master key as replaceable.
            if self.hs.config.mas.enabled:
                if not master_key_updatable_without_uia:
                    assert isinstance(self.auth, MasDelegatedAuth)
                    url = await self.auth.account_management_url()
                    url = f"{url}?action=org.matrix.cross_signing_reset"

                    # We use a dummy session ID as this isn't really a UIA flow, but we
                    # reuse the same API shape for better client compatibility.
                    raise InteractiveAuthIncompleteError(
                        "dummy",
                        {
                            "session": "dummy",
                            "flows": [
                                {"stages": ["m.oauth"]},
                                # The unstable name from MSC4312 should be supported until enough clients have adopted the stable (`m.oauth`) name:
                                {"stages": ["org.matrix.cross_signing_reset"]},
                            ],
                            "params": {
                                "m.oauth": {
                                    "url": url,
                                },
                                "org.matrix.cross_signing_reset": {
                                    "url": url,
                                },
                            },
                            "msg": "To reset your end-to-end encryption cross-signing "
                            f"identity, you first need to approve it at {url} and "
                            "then try again.",
                        },
                    )

            elif self.hs.config.experimental.msc3861.enabled:
                if not master_key_updatable_without_uia:
                    # If MSC3861 is enabled, we can assume self.auth is an instance of MSC3861DelegatedAuth
                    # We import lazily here because of the authlib requirement
                    from synapse.api.auth.msc3861_delegated import MSC3861DelegatedAuth

                    assert isinstance(self.auth, MSC3861DelegatedAuth)

                    uri = await self.auth.account_management_url()
                    if uri is not None:
                        url = f"{uri}?action=org.matrix.cross_signing_reset"
                    else:
                        url = await self.auth.issuer()

                    # We use a dummy session ID as this isn't really a UIA flow, but we
                    # reuse the same API shape for better client compatibility.
                    raise InteractiveAuthIncompleteError(
                        "dummy",
                        {
                            "session": "dummy",
                            "flows": [
                                {"stages": ["m.oauth"]},
                                # The unstable name from MSC4312 should be supported until enough clients have adopted the stable (`m.oauth`) name:
                                {"stages": ["org.matrix.cross_signing_reset"]},
                            ],
                            "params": {
                                "m.oauth": {
                                    "url": url,
                                },
                                "org.matrix.cross_signing_reset": {
                                    "url": url,
                                },
                            },
                            "msg": "To reset your end-to-end encryption cross-signing "
                            f"identity, you first need to approve it at {url} and "
                            "then try again.",
                        },
                    )

            else:
                # Without MSC3861, we require UIA.
                await self.auth_handler.validate_user_via_ui_auth(
                    requester,
                    request,
                    body,
                    "reset the device signing key on your account",
                    # Do not allow skipping of UIA auth.
                    can_skip_ui_auth=False,
                )

        result = await self.e2e_keys_handler.upload_signing_keys_for_user(user_id, body)
        return 200, result


class SignaturesUploadServlet(RestServlet):
    """
    POST /keys/signatures/upload HTTP/1.1
    Content-Type: application/json

    {
      "@alice:example.com": {
        "<device_id>": {
          "user_id": "<user_id>",
          "device_id": "<device_id>",
          "algorithms": [
            "m.olm.curve25519-aes-sha2",
            "m.megolm.v1.aes-sha2"
          ],
          "keys": {
            "<algorithm>:<device_id>": "<key_base64>",
          },
          "signatures": {
            "<signing_user_id>": {
              "<algorithm>:<signing_key_base64>": "<signature_base64>>"
            }
          }
        }
      }
    }
    """

    PATTERNS = client_patterns("/keys/signatures/upload$")

    def __init__(self, hs: "HomeServer"):
        super().__init__()
        self.auth = hs.get_auth()
        self.e2e_keys_handler = hs.get_e2e_keys_handler()

    async def on_POST(self, request: SynapseRequest) -> tuple[int, JsonDict]:
        requester = await self.auth.get_user_by_req(request, allow_guest=True)
        user_id = requester.user.to_string()
        body = parse_json_object_from_request(request)

        result = await self.e2e_keys_handler.upload_signatures_for_device_keys(
            user_id, body
        )
        return 200, result


def register_servlets(hs: "HomeServer", http_server: HttpServer) -> None:
    KeyUploadServlet(hs).register(http_server)
    KeyQueryServlet(hs).register(http_server)
    KeyChangesServlet(hs).register(http_server)
    OneTimeKeyServlet(hs).register(http_server)
    if hs.config.experimental.msc3983_appservice_otk_claims:
        UnstableOneTimeKeyServlet(hs).register(http_server)
    SigningKeyUploadServlet(hs).register(http_server)
    SignaturesUploadServlet(hs).register(http_server)



import logging

from synapse.types import JsonMapping

logger = logging.getLogger(__name__)


class AdminClientConfig:
    """Class to track various Synapse-specific admin-only client-impacting config options."""

    def __init__(self, account_data: JsonMapping | None):
        # Allow soft-failed events to be returned down `/sync` and other
        # client APIs. `io.element.synapse.soft_failed: true` is added to the
        # `unsigned` portion of the event to inform clients that the event
        # is soft-failed.
        self.return_soft_failed_events: bool = False
        self.return_policy_server_spammy_events: bool = False

        if account_data:
            self.return_soft_failed_events = account_data.get(
                "return_soft_failed_events", False
            )
            self.return_policy_server_spammy_events = account_data.get(
                "return_policy_server_spammy_events", self.return_soft_failed_events
            )



#
# This file is licensed under the Affero General Public License (AGPL) version 3.
#
# Copyright 2016-2021 The Matrix.org Foundation C.I.C.
# Copyright (C) 2023 New Vector, Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# See the GNU Affero General Public License for more details:
# <https://www.gnu.org/licenses/agpl-3.0.html>.
#
#

import logging
from typing import TYPE_CHECKING

from signedjson.key import decode_verify_key_bytes
from unpaddedbase64 import decode_base64

from synapse.api.errors import SynapseError
from synapse.crypto.keyring import VerifyJsonRequest
from synapse.events import EventBase
from synapse.types.handlers.policy_server import RECOMMENDATION_OK
from synapse.util.stringutils import parse_and_validate_server_name

if TYPE_CHECKING:
    from synapse.server import HomeServer

logger = logging.getLogger(__name__)

POLICY_SERVER_EVENT_TYPE = "org.matrix.msc4284.policy"
POLICY_SERVER_KEY_ID = "ed25519:policy_server"


class RoomPolicyHandler:
    def __init__(self, hs: "HomeServer"):
        self._hs = hs
        self._store = hs.get_datastores().main
        self._storage_controllers = hs.get_storage_controllers()
        self._event_auth_handler = hs.get_event_auth_handler()
        self._federation_client = hs.get_federation_client()

    async def is_event_allowed(self, event: EventBase) -> bool:
        """Check if the given event is allowed in the room by the policy server.

        Note: This will *always* return True if the room's policy server is Synapse
        itself. This is because Synapse can't be a policy server (currently).

        If no policy server is configured in the room, this returns True. Similarly, if
        the policy server is invalid in any way (not joined, not a server, etc), this
        returns True.

        If a valid and contactable policy server is configured in the room, this returns
        True if that server suggests the event is not spammy, and False otherwise.

        Args:
            event: The event to check. This should be a fully-formed PDU.

        Returns:
            bool: True if the event is allowed in the room, False otherwise.
        """
        if event.type == POLICY_SERVER_EVENT_TYPE and event.state_key is not None:
            return True  # always allow policy server change events

        policy_event = await self._storage_controllers.state.get_current_state_event(
            event.room_id, POLICY_SERVER_EVENT_TYPE, ""
        )
        if not policy_event:
            return True  # no policy server == default allow

        policy_server = policy_event.content.get("via", "")
        if policy_server is None or not isinstance(policy_server, str):
            return True  # no policy server == default allow

        if policy_server == self._hs.hostname:
            return True  # Synapse itself can't be a policy server (currently)

        try:
            parse_and_validate_server_name(policy_server)
        except ValueError:
            return True  # invalid policy server == default allow

        is_in_room = await self._event_auth_handler.is_host_in_room(
            event.room_id, policy_server
        )
        if not is_in_room:
            return True  # policy server not in room == default allow

        # Check if the event has been signed with the public key in the policy server state event.
        # If it is, we can save an HTTP hit.
        # We actually want to get the policy server state event BEFORE THE EVENT rather than
        # the current state value, else changing the public key will cause all of these checks to fail.
        # However, if we are checking outlier events (which we will due to is_event_allowed being called
        # near the edges at _check_sigs_and_hash) we won't know the state before the event, so the
        # only safe option is to use the current state
        public_key = policy_event.content.get("public_key", None)
        if public_key is not None and isinstance(public_key, str):
            valid = await self._verify_policy_server_signature(
                event, policy_server, public_key
            )
            if valid:
                return True
            # fallthrough to hit /check manually

        # At this point, the server appears valid and is in the room, so ask it to check
        # the event.
        recommendation = await self._federation_client.get_pdu_policy_recommendation(
            policy_server, event
        )
        if recommendation != RECOMMENDATION_OK:
            return False

        return True  # default allow

    async def _verify_policy_server_signature(
        self, event: EventBase, policy_server: str, public_key: str
    ) -> bool:
        # check the event is signed with this (via, public_key).
        verify_json_req = VerifyJsonRequest.from_event(policy_server, event, 0)
        try:
            key_bytes = decode_base64(public_key)
            verify_key = decode_verify_key_bytes(POLICY_SERVER_KEY_ID, key_bytes)
            # We would normally use KeyRing.verify_event_for_server but we can't here as we don't
            # want to fetch the server key, and instead want to use the public key in the state event.
            await self._hs.get_keyring().process_json(verify_key, verify_json_req)
            # if the event is correctly signed by the public key in the policy server state event = Allow
            return True
        except Exception as ex:
            logger.warning(
                "failed to verify event using public key in policy server event: %s", ex
            )
        return False

    async def ask_policy_server_to_sign_event(
        self, event: EventBase, verify: bool = False
    ) -> None:
        """Ask the policy server to sign this event. The signature is added to the event signatures block.

        Does nothing if there is no policy server state event in the room. If the policy server
        refuses to sign the event (as it's marked as spam) does nothing.

        Args:
            event: The event to sign
            verify: If True, verify that the signature is correctly signed by the public_key in the
            policy server state event.
        Raises:
            if verify=True and the policy server signed the event with an invalid signature. Does
            not raise if the policy server refuses to sign the event.
        """
        policy_event = await self._storage_controllers.state.get_current_state_event(
            event.room_id, POLICY_SERVER_EVENT_TYPE, ""
        )
        if not policy_event:
            return
        policy_server = policy_event.content.get("via", None)
        if policy_server is None or not isinstance(policy_server, str):
            return
        # Only ask to sign events if the policy state event has a public_key (so they can be subsequently verified)
        public_key = policy_event.content.get("public_key", None)
        if public_key is None or not isinstance(public_key, str):
            return

        # Ask the policy server to sign this event.
        # We set a smallish timeout here as we don't want to block event sending too long.
        signature = await self._federation_client.ask_policy_server_to_sign_event(
            policy_server,
            event,
            timeout=3000,
        )
        if (
            # the policy server returns {} if it refuses to sign the event.
            signature and len(signature) > 0
        ):
            event.signatures.update(signature)
            if verify:
                is_valid = await self._verify_policy_server_signature(
                    event, policy_server, public_key
                )
                if not is_valid:
                    raise SynapseError(
                        500,
                        f"policy server {policy_server} failed to sign event correctly",
                    )



#
# This file is licensed under the Affero General Public License (AGPL) version 3.
#
# Copyright 2020 The Matrix.org Foundation C.I.C.
# Copyright 2014-2016 OpenMarket Ltd
# Copyright (C) 2023 New Vector, Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# See the GNU Affero General Public License for more details:
# <https://www.gnu.org/licenses/agpl-3.0.html>.
#
# Originally licensed under the Apache License, Version 2.0:
# <http://www.apache.org/licenses/LICENSE-2.0>.
#
# [This file includes modifications made by New Vector Limited]
#
#
import collections.abc
import logging
from typing import (
    TYPE_CHECKING,
    Any,
    Collection,
    Iterable,
    Mapping,
    MutableMapping,
    TypeVar,
    cast,
    overload,
)

import attr

from synapse.api.constants import EventContentFields, EventTypes, Membership
from synapse.api.errors import NotFoundError, UnsupportedRoomVersionError
from synapse.api.room_versions import KNOWN_ROOM_VERSIONS, RoomVersion
from synapse.events import EventBase
from synapse.events.snapshot import EventContext
from synapse.logging.opentracing import trace
from synapse.replication.tcp.streams import UnPartialStatedEventStream
from synapse.replication.tcp.streams.partial_state import UnPartialStatedEventStreamRow
from synapse.storage._base import SQLBaseStore
from synapse.storage.database import (
    DatabasePool,
    LoggingDatabaseConnection,
    LoggingTransaction,
    make_in_list_sql_clause,
)
from synapse.storage.databases.main.events_worker import EventsWorkerStore
from synapse.storage.databases.main.roommember import RoomMemberWorkerStore
from synapse.types import JsonDict, JsonMapping, StateKey, StateMap, StrCollection
from synapse.types.state import StateFilter
from synapse.util.caches import intern_string
from synapse.util.caches.descriptors import cached, cachedList
from synapse.util.cancellation import cancellable
from synapse.util.iterutils import batch_iter

if TYPE_CHECKING:
    from synapse.server import HomeServer

logger = logging.getLogger(__name__)

_T = TypeVar("_T")

MAX_STATE_DELTA_HOPS = 100


# Freeze so it's immutable and we can use it as a cache value
@attr.s(slots=True, frozen=True, auto_attribs=True)
class Sentinel:
    pass


ROOM_UNKNOWN_SENTINEL = Sentinel()


@attr.s(slots=True, frozen=True, auto_attribs=True)
class EventMetadata:
    """Returned by `get_metadata_for_events`"""

    room_id: str
    event_type: str
    state_key: str | None
    rejection_reason: str | None


def _retrieve_and_check_room_version(room_id: str, room_version_id: str) -> RoomVersion:
    v = KNOWN_ROOM_VERSIONS.get(room_version_id)
    if not v:
        raise UnsupportedRoomVersionError(
            "Room %s uses a room version %s which is no longer supported"
            % (room_id, room_version_id)
        )
    return v


# this inherits from EventsWorkerStore because it calls self.get_events
class StateGroupWorkerStore(EventsWorkerStore, SQLBaseStore):
    """The parts of StateGroupStore that can be called from workers."""

    def __init__(
        self,
        database: DatabasePool,
        db_conn: LoggingDatabaseConnection,
        hs: "HomeServer",
    ):
        super().__init__(database, db_conn, hs)
        self._instance_name: str = hs.get_instance_name()

    def process_replication_rows(
        self,
        stream_name: str,
        instance_name: str,
        token: int,
        rows: Iterable[Any],
    ) -> None:
        if stream_name == UnPartialStatedEventStream.NAME:
            for row in rows:
                assert isinstance(row, UnPartialStatedEventStreamRow)
                self._get_state_group_for_event.invalidate((row.event_id,))
                self.is_partial_state_event.invalidate((row.event_id,))

        super().process_replication_rows(stream_name, instance_name, token, rows)

    async def get_room_version(self, room_id: str) -> RoomVersion:
        """Get the room_version of a given room
        Raises:
            NotFoundError: if the room is unknown
            UnsupportedRoomVersionError: if the room uses an unknown room version.
                Typically this happens if support for the room's version has been
                removed from Synapse.
        """
        room_version_id = await self.get_room_version_id(room_id)
        return _retrieve_and_check_room_version(room_id, room_version_id)

    def get_room_version_txn(
        self, txn: LoggingTransaction, room_id: str
    ) -> RoomVersion:
        """Get the room_version of a given room
        Args:
            txn: Transaction object
            room_id: The room_id of the room you are trying to get the version for
        Raises:
            NotFoundError: if the room is unknown
            UnsupportedRoomVersionError: if the room uses an unknown room version.
                Typically this happens if support for the room's version has been
                removed from Synapse.
        """
        room_version_id = self.get_room_version_id_txn(txn, room_id)
        return _retrieve_and_check_room_version(room_id, room_version_id)

    @cached(max_entries=10000)
    async def get_room_version_id(self, room_id: str) -> str:
        """Get the room_version of a given room
        Raises:
            NotFoundError: if the room is unknown
        """
        return await self.db_pool.runInteraction(
            "get_room_version_id_txn",
            self.get_room_version_id_txn,
            room_id,
        )

    def get_room_version_id_txn(self, txn: LoggingTransaction, room_id: str) -> str:
        """Get the room_version of a given room
        Args:
            txn: Transaction object
            room_id: The room_id of the room you are trying to get the version for
        Raises:
            NotFoundError: if the room is unknown
        """

        # We really should have an entry in the rooms table for every room we
        # care about, but let's be a bit paranoid.
        room_version = self.db_pool.simple_select_one_onecol_txn(
            txn,
            table="rooms",
            keyvalues={"room_id": room_id},
            retcol="room_version",
            allow_none=True,
        )

        if room_version is None:
            raise NotFoundError("Could not find room_version for %s" % (room_id,))

        return room_version

    @trace
    async def get_metadata_for_events(
        self, event_ids: Collection[str]
    ) -> dict[str, EventMetadata]:
        """Get some metadata (room_id, type, state_key) for the given events.

        This method is a faster alternative than fetching the full events from
        the DB, and should be used when the full event is not needed.

        Returns metadata for rejected and redacted events. Events that have not
        been persisted are omitted from the returned dict.
        """

        def get_metadata_for_events_txn(
            txn: LoggingTransaction,
            batch_ids: Collection[str],
        ) -> dict[str, EventMetadata]:
            clause, args = make_in_list_sql_clause(
                self.database_engine, "e.event_id", batch_ids
            )

            sql = f"""
                SELECT e.event_id, e.room_id, e.type, se.state_key, r.reason
                FROM events AS e
                LEFT JOIN state_events se USING (event_id)
                LEFT JOIN rejections r USING (event_id)
                WHERE {clause}
            """

            txn.execute(sql, args)
            return {
                event_id: EventMetadata(
                    room_id=room_id,
                    event_type=event_type,
                    state_key=state_key,
                    rejection_reason=rejection_reason,
                )
                for event_id, room_id, event_type, state_key, rejection_reason in txn
            }

        result_map: dict[str, EventMetadata] = {}
        for batch_ids in batch_iter(event_ids, 1000):
            result_map.update(
                await self.db_pool.runInteraction(
                    "get_metadata_for_events",
                    get_metadata_for_events_txn,
                    batch_ids=batch_ids,
                )
            )

        return result_map

    async def get_room_predecessor(self, room_id: str) -> JsonMapping | None:
        """Get the predecessor of an upgraded room if it exists.
        Otherwise return None.

        Args:
            room_id: The room ID.

        Returns:
            A dictionary containing the structure of the predecessor
            field from the room's create event. The structure is subject to other servers,
            but it is expected to be:
                * room_id (str): The room ID of the predecessor room
                * event_id (str): The ID of the tombstone event in the predecessor room

            None if a predecessor key is not found, or is not a dictionary.

        Raises:
            NotFoundError if the given room is unknown
        """
        # Retrieve the room's create event
        create_event = await self.get_create_event_for_room(room_id)

        # Retrieve the predecessor key of the create event
        predecessor = create_event.content.get("predecessor", None)

        # Ensure the key is a dictionary
        if not isinstance(predecessor, collections.abc.Mapping):
            return None

        # The keys must be strings since the data is JSON.
        return predecessor

    async def get_create_event_for_room(self, room_id: str) -> EventBase:
        """Get the create state event for a room.

        Args:
            room_id: The room ID.

        Returns:
            The room creation event.

        Raises:
            NotFoundError if the room is unknown
        """
        state_ids = await self.get_partial_current_state_ids(room_id)

        if not state_ids:
            raise NotFoundError(f"Current state for room {room_id} is empty")

        create_id = state_ids.get((EventTypes.Create, ""))

        # If we can't find the create event, assume we've hit a dead end
        if not create_id:
            raise NotFoundError(f"No create event in current state for room {room_id}")

        # Retrieve the room's create event and return
        create_event = await self.get_event(create_id)
        return create_event

    @cached(max_entries=10000)
    async def get_room_type(self, room_id: str) -> str | None | Sentinel:
        """Fetch room type for given room.

        Since this function is cached, any missing values would be cached as
        `None`. In order to distinguish between an unencrypted room that has
        `None` encryption and a room that is unknown to the server where we
        might want to omit the value (which would make it cached as `None`),
        instead we use the sentinel value `ROOM_UNKNOWN_SENTINEL`.
        """

        try:
            create_event = await self.get_create_event_for_room(room_id)
            return create_event.content.get(EventContentFields.ROOM_TYPE)
        except NotFoundError:
            # We use the sentinel value to distinguish between `None` which is a
            # valid room type and a room that is unknown to the server so the value
            # is just unset.
            return ROOM_UNKNOWN_SENTINEL

    @cachedList(cached_method_name="get_room_type", list_name="room_ids")
    async def bulk_get_room_type(
        self, room_ids: set[str]
    ) -> Mapping[str, str | None | Sentinel]:
        """
        Bulk fetch room types for the given rooms (via current state).

        Since this function is cached, any missing values would be cached as `None`. In
        order to distinguish between an unencrypted room that has `None` encryption and
        a room that is unknown to the server where we might want to omit the value
        (which would make it cached as `None`), instead we use the sentinel value
        `ROOM_UNKNOWN_SENTINEL`.

        Returns:
            A mapping from room ID to the room's type (`None` is a valid room type).
            Rooms unknown to this server will return `ROOM_UNKNOWN_SENTINEL`.
        """

        def txn(
            txn: LoggingTransaction,
        ) -> MutableMapping[str, str | None | Sentinel]:
            clause, args = make_in_list_sql_clause(
                txn.database_engine, "room_id", room_ids
            )

            # We can't rely on `room_stats_state.room_type` if the server has left the
            # room because the `room_id` will still be in the table but everything will
            # be set to `None` but `None` is a valid room type value. We join against
            # the `room_stats_current` table which keeps track of the
            # `current_state_events` count (and a proxy value `local_users_in_room`
            # which can used to assume the server is participating in the room and has
            # current state) to ensure that the data in `room_stats_state` is up-to-date
            # with the current state.
            #
            # FIXME: Use `room_stats_current.current_state_events` instead of
            # `room_stats_current.local_users_in_room` once
            # https://github.com/element-hq/synapse/issues/17457 is fixed.
            sql = f"""
                SELECT room_id, room_type
                FROM room_stats_state
                INNER JOIN room_stats_current USING (room_id)
                WHERE
                    {clause}
                    AND local_users_in_room > 0
            """

            txn.execute(sql, args)

            room_id_to_type_map = {}
            for row in txn:
                room_id_to_type_map[row[0]] = row[1]

            return room_id_to_type_map

        results = await self.db_pool.runInteraction(
            "bulk_get_room_type",
            txn,
        )

        # If we haven't updated `room_stats_state` with the room yet, query the
        # create events directly. This should happen only rarely so we don't
        # mind if we do this in a loop.
        for room_id in room_ids - results.keys():
            try:
                create_event = await self.get_create_event_for_room(room_id)
                room_type = create_event.content.get(EventContentFields.ROOM_TYPE)
                results[room_id] = room_type
            except NotFoundError:
                # We use the sentinel value to distinguish between `None` which is a
                # valid room type and a room that is unknown to the server so the value
                # is just unset.
                results[room_id] = ROOM_UNKNOWN_SENTINEL

        return results

    @cached(max_entries=10000)
    async def get_room_encryption(self, room_id: str) -> str | None:
        raise NotImplementedError()

    @cachedList(cached_method_name="get_room_encryption", list_name="room_ids")
    async def bulk_get_room_encryption(
        self, room_ids: set[str]
    ) -> Mapping[str, str | None | Sentinel]:
        """
        Bulk fetch room encryption for the given rooms (via current state).

        Since this function is cached, any missing values would be cached as `None`. In
        order to distinguish between an unencrypted room that has `None` encryption and
        a room that is unknown to the server where we might want to omit the value
        (which would make it cached as `None`), instead we use the sentinel value
        `ROOM_UNKNOWN_SENTINEL`.

        Returns:
            A mapping from room ID to the room's encryption algorithm if the room is
            encrypted, otherwise `None`. Rooms unknown to this server will return
            `ROOM_UNKNOWN_SENTINEL`.
        """

        def txn(
            txn: LoggingTransaction,
        ) -> MutableMapping[str, str | None | Sentinel]:
            clause, args = make_in_list_sql_clause(
                txn.database_engine, "room_id", room_ids
            )

            # We can't rely on `room_stats_state.encryption` if the server has left the
            # room because the `room_id` will still be in the table but everything will
            # be set to `None` but `None` is a valid encryption value. We join against
            # the `room_stats_current` table which keeps track of the
            # `current_state_events` count (and a proxy value `local_users_in_room`
            # which can used to assume the server is participating in the room and has
            # current state) to ensure that the data in `room_stats_state` is up-to-date
            # with the current state.
            #
            # FIXME: Use `room_stats_current.current_state_events` instead of
            # `room_stats_current.local_users_in_room` once
            # https://github.com/element-hq/synapse/issues/17457 is fixed.
            sql = f"""
                SELECT room_id, encryption
                FROM room_stats_state
                INNER JOIN room_stats_current USING (room_id)
                WHERE
                    {clause}
                    AND local_users_in_room > 0
            """

            txn.execute(sql, args)

            room_id_to_encryption_map = {}
            for row in txn:
                room_id_to_encryption_map[row[0]] = row[1]

            return room_id_to_encryption_map

        results = await self.db_pool.runInteraction(
            "bulk_get_room_encryption",
            txn,
        )

        # If we haven't updated `room_stats_state` with the room yet, query the state
        # directly. This should happen only rarely so we don't mind if we do this in a
        # loop.
        encryption_event_ids: list[str] = []
        for room_id in room_ids - results.keys():
            state_map = await self.get_partial_filtered_current_state_ids(
                room_id,
                state_filter=StateFilter.from_types(
                    [
                        (EventTypes.Create, ""),
                        (EventTypes.RoomEncryption, ""),
                    ]
                ),
            )
            # We can use the create event as a canary to tell whether the server has
            # seen the room before
            create_event_id = state_map.get((EventTypes.Create, ""))
            encryption_event_id = state_map.get((EventTypes.RoomEncryption, ""))

            if create_event_id is None:
                # We use the sentinel value to distinguish between `None` which is a
                # valid room type and a room that is unknown to the server so the value
                # is just unset.
                results[room_id] = ROOM_UNKNOWN_SENTINEL
                continue

            if encryption_event_id is None:
                results[room_id] = None
            else:
                encryption_event_ids.append(encryption_event_id)

        encryption_event_map = await self.get_events(encryption_event_ids)

        for encryption_event_id in encryption_event_ids:
            encryption_event = encryption_event_map.get(encryption_event_id)
            # If the curent state says there is an encryption event, we should have it
            # in the database.
            assert encryption_event is not None

            results[encryption_event.room_id] = encryption_event.content.get(
                EventContentFields.ENCRYPTION_ALGORITHM
            )

        return results

    @cached(max_entries=100000, iterable=True)
    async def get_partial_current_state_ids(self, room_id: str) -> StateMap[str]:
        """Get the current state event ids for a room based on the
        current_state_events table.

        This may be the partial state if we're lazy joining the room.

        Args:
            room_id: The room to get the state IDs of.

        Returns:
            The current state of the room.
        """

        def _get_current_state_ids_txn(txn: LoggingTransaction) -> StateMap[str]:
            txn.execute(
                """SELECT type, state_key, event_id FROM current_state_events
                WHERE room_id = ?
                """,
                (room_id,),
            )

            return {(intern_string(r[0]), intern_string(r[1])): r[2] for r in txn}

        return await self.db_pool.runInteraction(
            "get_partial_current_state_ids", _get_current_state_ids_txn
        )

    async def check_if_events_in_current_state(
        self, event_ids: StrCollection
    ) -> frozenset[str]:
        """Checks and returns which of the given events is part of the current state."""
        rows = await self.db_pool.simple_select_many_batch(
            table="current_state_events",
            column="event_id",
            iterable=event_ids,
            retcols=("event_id",),
            desc="check_if_events_in_current_state",
        )

        return frozenset(event_id for (event_id,) in rows)

    # FIXME: how should this be cached?
    @cancellable
    async def get_partial_filtered_current_state_ids(
        self, room_id: str, state_filter: StateFilter | None = None
    ) -> StateMap[str]:
        """Get the current state event of a given type for a room based on the
        current_state_events table.  This may not be as up-to-date as the result
        of doing a fresh state resolution as per state_handler.get_current_state

        This may be the partial state if we're lazy joining the room.

        Args:
            room_id
            state_filter: The state filter used to fetch state
                from the database.

        Returns:
            Map from type/state_key to event ID.
        """
        if state_filter is None:
            state_filter = StateFilter.all()

        where_clause, where_args = (state_filter).make_sql_filter_clause()

        if not where_clause:
            # We delegate to the cached version
            return await self.get_partial_current_state_ids(room_id)

        def _get_filtered_current_state_ids_txn(
            txn: LoggingTransaction,
        ) -> StateMap[str]:
            results = StateMapWrapper(state_filter=state_filter)

            sql = """
                SELECT type, state_key, event_id FROM current_state_events
                WHERE room_id = ?
            """

            if where_clause:
                sql += " AND (%s)" % (where_clause,)

            args = [room_id]
            args.extend(where_args)
            txn.execute(sql, args)
            for row in txn:
                typ, state_key, event_id = row
                key = (intern_string(typ), intern_string(state_key))
                results[key] = event_id

            return results

        return await self.db_pool.runInteraction(
            "get_filtered_current_state_ids", _get_filtered_current_state_ids_txn
        )

    @cached(max_entries=50000)
    async def _get_state_group_for_event(self, event_id: str) -> int | None:
        return await self.db_pool.simple_select_one_onecol(
            table="event_to_state_groups",
            keyvalues={"event_id": event_id},
            retcol="state_group",
            allow_none=True,
            desc="_get_state_group_for_event",
        )

    @cachedList(
        cached_method_name="_get_state_group_for_event",
        list_name="event_ids",
        num_args=1,
    )
    async def _get_state_group_for_events(
        self, event_ids: Collection[str]
    ) -> Mapping[str, int]:
        """Returns mapping event_id -> state_group.

        Raises:
             RuntimeError if the state is unknown at any of the given events
        """
        rows = cast(
            list[tuple[str, int]],
            await self.db_pool.simple_select_many_batch(
                table="event_to_state_groups",
                column="event_id",
                iterable=event_ids,
                keyvalues={},
                retcols=("event_id", "state_group"),
                desc="_get_state_group_for_events",
            ),
        )

        res = dict(rows)
        for e in event_ids:
            if e not in res:
                raise RuntimeError("No state group for unknown or outlier event %s" % e)
        return res

    async def get_referenced_state_groups(
        self, state_groups: Iterable[int]
    ) -> set[int]:
        """Check if the state groups are referenced by events.

        Args:
            state_groups

        Returns:
            The subset of state groups that are referenced.
        """

        rows = cast(
            list[tuple[int]],
            await self.db_pool.simple_select_many_batch(
                table="event_to_state_groups",
                column="state_group",
                iterable=state_groups,
                keyvalues={},
                retcols=("DISTINCT state_group",),
                desc="get_referenced_state_groups",
            ),
        )

        return {row[0] for row in rows}

    async def update_state_for_partial_state_event(
        self,
        event: EventBase,
        context: EventContext,
    ) -> None:
        """Update the state group for a partial state event"""
        async with (
            self._un_partial_stated_events_stream_id_gen.get_next() as un_partial_state_event_stream_id
        ):
            await self.db_pool.runInteraction(
                "update_state_for_partial_state_event",
                self._update_state_for_partial_state_event_txn,
                event,
                context,
                un_partial_state_event_stream_id,
            )

    def _update_state_for_partial_state_event_txn(
        self,
        txn: LoggingTransaction,
        event: EventBase,
        context: EventContext,
        un_partial_state_event_stream_id: int,
    ) -> None:
        # we shouldn't have any outliers here
        assert not event.internal_metadata.is_outlier()

        # anything that was rejected should have the same state as its
        # predecessor.
        if context.rejected:
            state_group = context.state_group_before_event
        else:
            state_group = context.state_group

        self.db_pool.simple_update_txn(
            txn,
            table="event_to_state_groups",
            keyvalues={"event_id": event.event_id},
            updatevalues={"state_group": state_group},
        )

        # the event may now be rejected where it was not before, or vice versa,
        # in which case we need to update the rejected flags.
        rejection_status_changed = bool(context.rejected) != (
            event.rejected_reason is not None
        )
        if rejection_status_changed:
            self.mark_event_rejected_txn(txn, event.event_id, context.rejected)

        self.db_pool.simple_delete_one_txn(
            txn,
            table="partial_state_events",
            keyvalues={"event_id": event.event_id},
        )

        txn.call_after(self.is_partial_state_event.invalidate, (event.event_id,))
        txn.call_after(
            self._get_state_group_for_event.prefill,
            (event.event_id,),
            state_group,
        )

        self.db_pool.simple_insert_txn(
            txn,
            "un_partial_stated_event_stream",
            {
                "stream_id": un_partial_state_event_stream_id,
                "instance_name": self._instance_name,
                "event_id": event.event_id,
                "rejection_status_changed": rejection_status_changed,
            },
        )
        txn.call_after(self.hs.get_notifier().on_new_replication_data)


class MainStateBackgroundUpdateStore(RoomMemberWorkerStore):
    CURRENT_STATE_INDEX_UPDATE_NAME = "current_state_members_idx"
    EVENT_STATE_GROUP_INDEX_UPDATE_NAME = "event_to_state_groups_sg_index"
    DELETE_CURRENT_STATE_UPDATE_NAME = "delete_old_current_state_events"
    MEMBERS_CURRENT_STATE_UPDATE_NAME = "current_state_events_members_room_index"

    def __init__(
        self,
        database: DatabasePool,
        db_conn: LoggingDatabaseConnection,
        hs: "HomeServer",
    ):
        super().__init__(database, db_conn, hs)

        self.server_name: str = hs.hostname

        self.db_pool.updates.register_background_index_update(
            self.CURRENT_STATE_INDEX_UPDATE_NAME,
            index_name="current_state_events_member_index",
            table="current_state_events",
            columns=["state_key"],
            where_clause="type='m.room.member'",
        )
        self.db_pool.updates.register_background_index_update(
            self.EVENT_STATE_GROUP_INDEX_UPDATE_NAME,
            index_name="event_to_state_groups_sg_index",
            table="event_to_state_groups",
            columns=["state_group"],
        )
        self.db_pool.updates.register_background_update_handler(
            self.DELETE_CURRENT_STATE_UPDATE_NAME,
            self._background_remove_left_rooms,
        )
        self.db_pool.updates.register_background_index_update(
            self.MEMBERS_CURRENT_STATE_UPDATE_NAME,
            index_name="current_state_events_members_room_index",
            table="current_state_events",
            columns=["room_id", "membership"],
            where_clause="type='m.room.member'",
        )

    async def _background_remove_left_rooms(
        self, progress: JsonDict, batch_size: int
    ) -> int:
        """Background update to delete rows from `current_state_events` and
        `event_forward_extremities` tables of rooms that the server is no
        longer joined to.
        """

        last_room_id = progress.get("last_room_id", "")

        def _background_remove_left_rooms_txn(
            txn: LoggingTransaction,
        ) -> tuple[bool, set[str]]:
            # get a batch of room ids to consider
            sql = """
                SELECT DISTINCT room_id FROM current_state_events
                WHERE room_id > ? ORDER BY room_id LIMIT ?
            """

            txn.execute(sql, (last_room_id, batch_size))
            room_ids = [row[0] for row in txn]
            if not room_ids:
                return True, set()

            ###########################################################################
            #
            # exclude rooms where we have active members

            sql = """
                SELECT room_id
                FROM local_current_membership
                WHERE
                    room_id > ? AND room_id <= ?
                    AND membership = 'join'
                GROUP BY room_id
            """

            txn.execute(sql, (last_room_id, room_ids[-1]))
            joined_room_ids = {row[0] for row in txn}
            to_delete = set(room_ids) - joined_room_ids

            ###########################################################################
            #
            # exclude rooms which we are in the process of constructing; these otherwise
            # qualify as "rooms with no local users", and would have their
            # forward extremities cleaned up.

            # the following query will return a list of rooms which have forward
            # extremities that are *not* also the create event in the room - ie
            # those that are not being created currently.

            sql = """
                SELECT DISTINCT efe.room_id
                FROM event_forward_extremities efe
                LEFT JOIN current_state_events cse ON
                    cse.event_id = efe.event_id
                    AND cse.type = 'm.room.create'
                    AND cse.state_key = ''
                WHERE
                    cse.event_id IS NULL
                    AND efe.room_id > ? AND efe.room_id <= ?
            """

            txn.execute(sql, (last_room_id, room_ids[-1]))

            # build a set of those rooms within `to_delete` that do not appear in
            # the above, leaving us with the rooms in `to_delete` that *are* being
            # created.
            creating_rooms = to_delete.difference(row[0] for row in txn)
            logger.info("skipping rooms which are being created: %s", creating_rooms)

            # now remove the rooms being created from the list of those to delete.
            #
            # (we could have just taken the intersection of `to_delete` with the result
            # of the sql query, but it's useful to be able to log `creating_rooms`; and
            # having done so, it's quicker to remove the (few) creating rooms from
            # `to_delete` than it is to form the intersection with the (larger) list of
            # not-creating-rooms)

            to_delete -= creating_rooms

            ###########################################################################
            #
            # now clear the state for the rooms

            logger.info("Deleting current state left rooms: %r", to_delete)

            # First we get all users that we still think were joined to the
            # room. This is so that we can mark those device lists as
            # potentially stale, since there may have been a period where the
            # server didn't share a room with the remote user and therefore may
            # have missed any device updates.
            rows = cast(
                list[tuple[str]],
                self.db_pool.simple_select_many_txn(
                    txn,
                    table="current_state_events",
                    column="room_id",
                    iterable=to_delete,
                    keyvalues={
                        "type": EventTypes.Member,
                        "membership": Membership.JOIN,
                    },
                    retcols=("state_key",),
                ),
            )

            potentially_left_users = {row[0] for row in rows}

            # Now lets actually delete the rooms from the DB.
            self.db_pool.simple_delete_many_txn(
                txn,
                table="current_state_events",
                column="room_id",
                values=to_delete,
                keyvalues={},
            )

            self.db_pool.simple_delete_many_txn(
                txn,
                table="event_forward_extremities",
                column="room_id",
                values=to_delete,
                keyvalues={},
            )

            self.db_pool.updates._background_update_progress_txn(
                txn,
                self.DELETE_CURRENT_STATE_UPDATE_NAME,
                {"last_room_id": room_ids[-1]},
            )

            return False, potentially_left_users

        finished, potentially_left_users = await self.db_pool.runInteraction(
            "_background_remove_left_rooms", _background_remove_left_rooms_txn
        )

        if finished:
            await self.db_pool.updates._end_background_update(
                self.DELETE_CURRENT_STATE_UPDATE_NAME
            )

        # Now go and check if we still share a room with the remote users in
        # the deleted rooms. If not mark their device lists as stale.
        joined_users = await self.get_users_server_still_shares_room_with(
            potentially_left_users
        )

        for user_id in potentially_left_users - joined_users:
            await self.mark_remote_user_device_list_as_unsubscribed(user_id)  # type: ignore[attr-defined]

        return batch_size


class StateStore(StateGroupWorkerStore, MainStateBackgroundUpdateStore):
    """Keeps track of the state at a given event.

    This is done by the concept of `state groups`. Every event is a assigned
    a state group (identified by an arbitrary string), which references a
    collection of state events. The current state of an event is then the
    collection of state events referenced by the event's state group.

    Hence, every change in the current state causes a new state group to be
    generated. However, if no change happens (e.g., if we get a message event
    with only one parent it inherits the state group from its parent.)

    There are three tables:
      * `state_groups`: Stores group name, first event with in the group and
        room id.
      * `event_to_state_groups`: Maps events to state groups.
      * `state_groups_state`: Maps state group to state events.
    """

    def __init__(
        self,
        database: DatabasePool,
        db_conn: LoggingDatabaseConnection,
        hs: "HomeServer",
    ):
        super().__init__(database, db_conn, hs)


@attr.s(auto_attribs=True, slots=True)
class StateMapWrapper(dict[StateKey, str]):
    """A wrapper around a StateMap[str] to ensure that we only query for items
    that were not filtered out.

    This is to help prevent bugs where we filter out state but other bits of the
    code expect the state to be there.
    """

    state_filter: StateFilter

    def __getitem__(self, key: StateKey) -> str:
        if key not in self.state_filter:
            raise Exception("State map was filtered and doesn't include: %s", key)
        return super().__getitem__(key)

    @overload
    def get(self, key: StateKey, default: None = None, /) -> str | None: ...
    @overload
    def get(self, key: StateKey, default: str, /) -> str: ...
    @overload
    def get(self, key: StateKey, default: _T, /) -> str | _T: ...

    def get(self, key: StateKey, default: str | _T | None = None) -> str | _T | None:
        if key not in self.state_filter:
            raise Exception("State map was filtered and doesn't include: %s", key)
        return super().get(key, default)

    def __contains__(self, key: Any) -> bool:
        if key not in self.state_filter:
            raise Exception("State map was filtered and doesn't include: %s", key)

        return super().__contains__(key)



#
# This file is licensed under the Affero General Public License (AGPL) version 3.
#
# Copyright 2020 The Matrix.org Foundation C.I.C.
# Copyright 2015, 2016 OpenMarket Ltd
# Copyright (C) 2024 New Vector, Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# See the GNU Affero General Public License for more details:
# <https://www.gnu.org/licenses/agpl-3.0.html>.
#
# Originally licensed under the Apache License, Version 2.0:
# <http://www.apache.org/licenses/LICENSE-2.0>.
#
# [This file includes modifications made by New Vector Limited]
#
#

import logging
import re

from synapse.api.errors import Codes, cs_error
from synapse.http.server import (
    HttpServer,
    respond_with_json,
    respond_with_json_bytes,
    set_corp_headers,
    set_cors_headers,
)
from synapse.http.servlet import RestServlet, parse_integer, parse_string
from synapse.http.site import SynapseRequest
from synapse.media._base import (
    DEFAULT_MAX_TIMEOUT_MS,
    MAXIMUM_ALLOWED_MAX_TIMEOUT_MS,
    respond_404,
)
from synapse.media.media_repository import MediaRepository
from synapse.media.media_storage import MediaStorage
from synapse.media.thumbnailer import ThumbnailProvider
from synapse.server import HomeServer
from synapse.util.stringutils import parse_and_validate_server_name

logger = logging.getLogger(__name__)


class PreviewURLServlet(RestServlet):
    """
    Same as `GET /_matrix/media/r0/preview_url`, this endpoint provides a generic preview API
    for URLs which outputs Open Graph (https://ogp.me/) responses (with some Matrix
    specific additions).

    This does have trade-offs compared to other designs:

    * Pros:
      * Simple and flexible; can be used by any clients at any point
    * Cons:
      * If each homeserver provides one of these independently, all the homeservers in a
        room may needlessly DoS the target URI
      * The URL metadata must be stored somewhere, rather than just using Matrix
        itself to store the media.
      * Matrix cannot be used to distribute the metadata between homeservers.
    """

    PATTERNS = [re.compile(r"^/_matrix/client/v1/media/preview_url$")]

    def __init__(
        self,
        hs: "HomeServer",
        media_repo: "MediaRepository",
        media_storage: MediaStorage,
    ):
        super().__init__()

        self.auth = hs.get_auth()
        self.clock = hs.get_clock()
        self.media_repo = media_repo
        self.media_storage = media_storage
        assert self.media_repo.url_previewer is not None
        self.url_previewer = self.media_repo.url_previewer

    async def on_GET(self, request: SynapseRequest) -> None:
        requester = await self.auth.get_user_by_req(request)
        url = parse_string(request, "url", required=True)
        ts = parse_integer(request, "ts")
        if ts is None:
            ts = self.clock.time_msec()

        og = await self.url_previewer.preview(url, requester.user, ts)
        respond_with_json_bytes(request, 200, og, send_cors=True)


class MediaConfigResource(RestServlet):
    PATTERNS = [re.compile(r"^/_matrix/client/v1/media/config$")]

    def __init__(self, hs: "HomeServer"):
        super().__init__()
        config = hs.config
        self.clock = hs.get_clock()
        self.auth = hs.get_auth()
        self.limits_dict = {"m.upload.size": config.media.max_upload_size}
        self.media_repository_callbacks = hs.get_module_api_callbacks().media_repository

    async def on_GET(self, request: SynapseRequest) -> None:
        requester = await self.auth.get_user_by_req(request)
        user_specific_config = (
            await self.media_repository_callbacks.get_media_config_for_user(
                requester.user.to_string(),
            )
        )
        response = user_specific_config if user_specific_config else self.limits_dict
        respond_with_json(request, 200, response, send_cors=True)


class ThumbnailResource(RestServlet):
    PATTERNS = [
        re.compile(
            "/_matrix/client/v1/media/thumbnail/(?P<server_name>[^/]*)/(?P<media_id>[^/]*)$"
        )
    ]

    def __init__(
        self,
        hs: "HomeServer",
        media_repo: "MediaRepository",
        media_storage: MediaStorage,
    ):
        super().__init__()

        self.store = hs.get_datastores().main
        self.media_repo = media_repo
        self.media_storage = media_storage
        self.dynamic_thumbnails = hs.config.media.dynamic_thumbnails
        self._is_mine_server_name = hs.is_mine_server_name
        self._server_name = hs.hostname
        self.prevent_media_downloads_from = hs.config.media.prevent_media_downloads_from
        self.thumbnailer = ThumbnailProvider(hs, media_repo, media_storage)
        self.auth = hs.get_auth()

    async def on_GET(
        self, request: SynapseRequest, server_name: str, media_id: str
    ) -> None:
        # Validate the server name, raising if invalid
        parse_and_validate_server_name(server_name)
        await self.auth.get_user_by_req(request, allow_guest=True)

        set_cors_headers(request)
        set_corp_headers(request)
        width = parse_integer(request, "width", required=True)
        height = parse_integer(request, "height", required=True)
        method = parse_string(request, "method", "scale")
        # TODO Parse the Accept header to get an prioritised list of thumbnail types.
        m_type = "image/png"
        max_timeout_ms = parse_integer(
            request, "timeout_ms", default=DEFAULT_MAX_TIMEOUT_MS
        )
        max_timeout_ms = min(max_timeout_ms, MAXIMUM_ALLOWED_MAX_TIMEOUT_MS)

        if self._is_mine_server_name(server_name):
            if self.dynamic_thumbnails:
                await self.thumbnailer.select_or_generate_local_thumbnail(
                    request,
                    media_id,
                    width,
                    height,
                    method,
                    m_type,
                    max_timeout_ms,
                    False,
                )
            else:
                await self.thumbnailer.respond_local_thumbnail(
                    request,
                    media_id,
                    width,
                    height,
                    method,
                    m_type,
                    max_timeout_ms,
                    False,
                )
            self.media_repo.mark_recently_accessed(None, media_id)
        else:
            # Don't let users download media from configured domains, even if it
            # is already downloaded. This is Trust & Safety tooling to make some
            # media inaccessible to local users.
            # See `prevent_media_downloads_from` config docs for more info.
            if server_name in self.prevent_media_downloads_from:
                respond_404(request)
                return

            ip_address = request.getClientAddress().host
            remote_resp_function = (
                self.thumbnailer.select_or_generate_remote_thumbnail
                if self.dynamic_thumbnails
                else self.thumbnailer.respond_remote_thumbnail
            )
            await remote_resp_function(
                request,
                server_name,
                media_id,
                width,
                height,
                method,
                m_type,
                max_timeout_ms,
                ip_address,
                True,
            )
            self.media_repo.mark_recently_accessed(server_name, media_id)


class DownloadResource(RestServlet):
    PATTERNS = [
        re.compile(
            "/_matrix/client/v1/media/download/(?P<server_name>[^/]*)/(?P<media_id>[^/]*)(/(?P<file_name>[^/]*))?$"
        )
    ]

    def __init__(self, hs: "HomeServer", media_repo: "MediaRepository"):
        super().__init__()
        self.media_repo = media_repo
        self._is_mine_server_name = hs.is_mine_server_name
        self.auth = hs.get_auth()

    async def on_GET(
        self,
        request: SynapseRequest,
        server_name: str,
        media_id: str,
        file_name: str | None = None,
    ) -> None:
        # Validate the server name, raising if invalid
        parse_and_validate_server_name(server_name)

        requester = await self.auth.get_user_by_req(request, allow_guest=True)
        is_admin = await self.auth.is_server_admin(requester)
        bypass_quarantine = False
        if parse_string(request, "admin_unsafely_bypass_quarantine") == "true":
            if is_admin:
                logger.info("Admin bypassing quarantine for media download")
                bypass_quarantine = True
            else:
                respond_with_json(
                    request,
                    400,
                    cs_error(
                        "Must be a server admin to bypass quarantine",
                        code=Codes.UNKNOWN,
                    ),
                    send_cors=True,
                )

        set_cors_headers(request)
        set_corp_headers(request)
        request.setHeader(
            b"Content-Security-Policy",
            b"sandbox;"
            b" default-src 'none';"
            b" script-src 'none';"
            b" plugin-types application/pdf;"
            b" style-src 'unsafe-inline';"
            b" media-src 'self';"
            b" object-src 'self';",
        )
        # Limited non-standard form of CSP for IE11
        request.setHeader(b"X-Content-Security-Policy", b"sandbox;")
        request.setHeader(b"Referrer-Policy", b"no-referrer")
        max_timeout_ms = parse_integer(
            request, "timeout_ms", default=DEFAULT_MAX_TIMEOUT_MS
        )
        max_timeout_ms = min(max_timeout_ms, MAXIMUM_ALLOWED_MAX_TIMEOUT_MS)

        if self._is_mine_server_name(server_name):
            await self.media_repo.get_local_media(
                request,
                media_id,
                file_name,
                max_timeout_ms,
                bypass_quarantine=bypass_quarantine,
            )
        else:
            ip_address = request.getClientAddress().host
            await self.media_repo.get_remote_media(
                request,
                server_name,
                media_id,
                file_name,
                max_timeout_ms,
                ip_address,
                True,
                bypass_quarantine=bypass_quarantine,
            )


def register_servlets(hs: "HomeServer", http_server: HttpServer) -> None:
    media_repo = hs.get_media_repository()
    if hs.config.media.url_preview_enabled:
        PreviewURLServlet(hs, media_repo, media_repo.media_storage).register(
            http_server
        )
    MediaConfigResource(hs).register(http_server)
    ThumbnailResource(hs, media_repo, media_repo.media_storage).register(http_server)
    DownloadResource(hs, media_repo).register(http_server)
